;nyquist plug-in
;version 3
;type process
;categories "http://lv2plug.in/ns/lv2core/#DelayPlugin"
;name "Extend Silences..."
;action "Finding silences then inserting more..."
;info "by Steve Daulton (www.easyspacepro.com).\nReleased under GPL v2.\n"

;; ExtendSilence.ny by Steve Daulton June 2013.
;; Released under terms of the GNU General Public License version 2:
;; http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
;;
;; For information about writing and modifying Nyquist plug-ins:
;; http://wiki.audacityteam.org/wiki/Nyquist_Plug-ins_Reference

;control threshold "Threshold for silence (dB)" real "" -40 -100 0
;control min-dur "Min detected silence (seconds)" real "" 1 0.1 10.1
;control add-dur "Add silence duration (seconds)" real "" 5 0 10
;control add-sil-pc "Add % of detected silence duration" real "" 0 0 100
;control add-snd-pc "Add % of previous sound duration" real "" 0 0 100


;;; Constrain user input to sensible range.
(defmacro constrain (x xmin xmax)
  `(setq ,x (min ,xmax (max ,xmin ,x))))

(constrain threshold -1000 0)
(constrain min-dur 0.1 100)
(constrain add-dur 0 100)
(setq sil-ratio (/ (constrain add-sil-pc 0 1000) 100.0))
(setq snd-ratio (/ (constrain add-snd-pc 0 1000) 100.0))


;;; For stereo tracks, use maximum of L/R channels
(defun mono (s-in)
  (if (arrayp s-in)               ; if stereo,
      (s-max                      ; absolute max of channels.
        (snd-abs (aref s-in 0))
        (snd-abs (aref s-in 1)))
      (snd-abs s-in)))            ; else absolute sample values.


;;; Low sample rate copy of sound.
(defun s-lorate (s-in)
  (let ((step (round (/ *sound-srate* 100))))
    (snd-avg (mono s-in) step step OP-PEAK)))


;;; List of (time silences end)(duration of silence):
(defun get-list (sig thresh mindur)
  (let* ((sig (s-lorate sig))               ; get low samplerate copy.
         (lo-rate (snd-srate sig))          ; sample rate of sig.
         (thresh (db-to-linear thresh))     ; threshold dB to linear.
         (mindur (* mindur lo-rate))        ; min duration in samples.
         (sil-list '((0 0)))       ; initialise list.
         (sil-count 0))                     ; count 'silent' samples.
    ;; Loop through samples of low sample rate signal:
    (do ((val (snd-fetch sig)(snd-fetch sig))
         (count 0 (1+ count)))              ; sample counter.
        ((not val))                         ; end when sample value is NIL.
      ;; sil-count > 0. We are in a silence.
      (if (< val thresh)                    ; below threshold.
          (setq sil-count (1+ sil-count))
          ;; above threshold.
          (prog ()
            (when (>= sil-count mindur)
              (push (list (/ count lo-rate) (/ sil-count lo-rate)) sil-list))
            (setq sil-count 0))))           ; reset silence counter.
      ;; handle trailing silence.
      (when (>= sil-count mindur)
        (push (list (get-duration 1) (/ sil-count lo-rate)) sil-list))
      ;; return list of silences.
      sil-list))


;;; Add pause between each sound in sil-list:
(defun addpause (sig pause sil-list)
  (do ((n 0 (1+ n))
       (snd-pc-total 0)         ; running total of added '% of sound' added.
       (sil-pc-total 0)         ; running total of added '% of silence' added.
       (output (s-rest 0)))     ; initialise output.
      ((= n (1- (length sil-list))) output) 
    (let* ((start (first (nth n sil-list)))     ; start of section.
           (end (first (nth (1+ n) sil-list)))  ; end of section.
           (section-dur (- end start))          ; duration of section.
           (sil-len (second (nth n sil-list)))) ; length of silence
      ; running total of added '% of silence' added.
      (setq sil-pc-total (+ sil-pc-total (* sil-ratio sil-len)))
      (setq total-pauses        ; total of added pauses.
        (+ sil-pc-total snd-pc-total (* n pause)))
      ; total duration so far.
      (setq outdur (+ end total-pauses))
      (setf section (extract-abs start end sig))
      ;; Add next section of track to output.
      (setf output
        (extract-abs 0 outdur
          (sim
            output
            (at-abs (- outdur section-dur) (cue section)))))
      ;; Calculate length of sound in current section and
      ;; derive total percent of sound duration added in next gap.
      (let* ((silence-in-section (second (nth (1+ n) sil-list)))
             (snd-sel-len (* snd-ratio (- section-dur silence-in-section))))
        (setq snd-pc-total (+ snd-pc-total snd-sel-len))))))


(if (= (+ add-dur sil-ratio snd-ratio) 0) ; No silence to be added
    "Please check control settings.\nTotal silence to add is zero."
    ;; else
    (let ((time-list (get-list s threshold min-dur)))
      (cond
        ((> (length time-list) 1)  ; If silences found:
          ; Add final end time.
          (push (list (get-duration 1) 0) time-list)
          ; reverse list into chronological order.
          (setf time-list (reverse time-list))
          ; Expand silences.
          (multichan-expand #'addpause s add-dur time-list))
        (T ; Else, warn the user:
          (format nil "No silences found below ~a dB~%~
                  and longer than ~a seconds."
                  threshold
                  min-dur)))))
